/*
 *		Basic Direct3D program.
 *		This program shows you an example about 
 *		the typical Initialization and typical 
 *		termination for Direct3D.
 *
 *		First edition: August.12.2002
 */

#include "StdAfx.h"
#include <math.h>
#include <d3d8.h>
#include <d3dx8.h>
#include <mmsystem.h>
#include "GameMain.h"
#include "DeviceInformation.h"
#include "HighResTimer.h"
#include "Object3D.h"
 /*
 *		Globals.
 */
HWND					g_hWnd;
D3DPRESENT_PARAMETERS	g_d3dprm;
LPDIRECT3D8				g_lpD3D = NULL;
LPDIRECT3DDEVICE8		g_lpD3DDEV = NULL;

CDeviceInformation	*g_lpDevices = NULL;
D3DAdapterInfo		*g_Adapters = NULL;
DWORD				g_dwNumAdapters	= 0;
DWORD				g_dwAdapter		= 0L;
BOOL				g_bWindowed = FALSE;
BOOL				g_bUseDepthBuffer   = TRUE;
DWORD				g_dwMinDepthBits    = 16;
DWORD				g_dwMinStencilBits  = 0;
DWORD				g_dwWindowStyle;
RECT				g_rcWindowBounds;
RECT				g_rcWindowClient;
BOOLEAN				g_bReady = false;
D3DCAPS8			g_CurrentDeviceCaps;
D3DSURFACE_DESC		g_d3dsdBackBuffer;

D3DXMATRIX		g_matView;
D3DXMATRIX		g_matProj;
D3DXVECTOR3		g_vecLight;
D3DCOLOR		g_bgColor = D3DCOLOR_XRGB(BGC_R,BGC_G,BGC_B);

CHighResTimer	*g_pHighResTimer = NULL;

CObject3D		*g_pObject3D = NULL;
/*
 *		Constants.
 */
#define	CAMERA_Z	(-5.0f)

/*
 *		Prototypes.
 */
void	SetupLights();
HRESULT Initialize3DEnvironment(HWND hWnd);
void	SetDefaultRenderState(LPDIRECT3DDEVICE8 lpd3ddev);
void	SetDefaultTransform(LPDIRECT3DDEVICE8 lpd3ddev);

void	ReleaseD3D()
{
	InvalidateDeviceObjects();
	DeleteDeviceObjects();
	if (g_pObject3D != NULL){
		delete g_pObject3D;
		g_pObject3D = NULL;
	}
	if (g_pHighResTimer != NULL){
		delete g_pHighResTimer;
		g_pHighResTimer = NULL;
	}
	SAFE_RELEASE(g_lpD3DDEV);
	SAFE_RELEASE(g_lpD3D);
	if (g_lpDevices){
		delete	g_lpDevices;
		g_lpDevices = NULL;
	}
}

#define TITLE               "Direct3D Miss"

//-----------------------------------------------------------------------------
// Name: InitFail()
// sɌĂ΂֐
//-----------------------------------------------------------------------------
void	InitFail(HWND hWnd, LPCTSTR szError,...)
{
    char                        szBuff[128];
    va_list                     vl;

    va_start(vl, szError);
    vsprintf(szBuff, szError, vl);
    MessageBox(hWnd, szBuff, TITLE, MB_OK);
    ReleaseD3D();
	DestroyWindow(hWnd);
    va_end(vl);
    return;
}

void	InitD3D(HWND hWnd)
{
	g_pObject3D = new CObject3D();
	TRACE("DIRECT3D_VERSION %x",DIRECT3D_VERSION);
	g_hWnd = hWnd;
	g_lpD3D = Direct3DCreate8(D3D_SDK_VERSION);
	if (!g_lpD3D){
		InitFail(hWnd,"Failed to create Direct3D");
		return;
	}
	g_dwWindowStyle = GetWindowLong( hWnd, GWL_STYLE );
    GetWindowRect( hWnd, &g_rcWindowBounds );
    GetClientRect( hWnd, &g_rcWindowClient );

	g_lpDevices = new CDeviceInformation(g_lpD3D, g_bUseDepthBuffer,
									g_dwMinDepthBits, g_dwMinStencilBits,
									NULL);
    if ((NULL == g_lpDevices)||( NULL == (g_Adapters = g_lpDevices->getAdapters())))
    {
		InitFail(hWnd, "BuildDeviceList Failed.");
        return;
    }
	g_dwNumAdapters = g_lpDevices->getNumAdapters();
	g_dwAdapter = g_lpDevices->getCurrentAdapter();

	//	Create device with hardware geometry engine.
	if (FAILED(::Initialize3DEnvironment(hWnd))){
		InitFail(hWnd, "CreateDevice Failed.");
		return;
	}
	if (FAILED(g_lpD3DDEV->GetDeviceCaps(&g_CurrentDeviceCaps))){
		InitFail(hWnd, "GetDeviceCaps Failed.");
		return;
	}
	g_pHighResTimer = new CHighResTimer();
}


//-----------------------------------------------------------------------------
// Name: Initialize3DEnvironment()
// Desc:
//-----------------------------------------------------------------------------
HRESULT Initialize3DEnvironment(HWND hWnd)
{
    HRESULT hr;

    D3DAdapterInfo* pAdapterInfo = &g_Adapters[g_dwAdapter];
    D3DDeviceInfo*  pDeviceInfo  = &pAdapterInfo->devices[pAdapterInfo->dwCurrentDevice];
    D3DModeInfo*    pModeInfo    = &pDeviceInfo->modes[pDeviceInfo->dwCurrentMode];

    // Prepare window for possible windowed/fullscreen change

    // Set up the presentation parameters
    ZeroMemory( &g_d3dprm, sizeof(g_d3dprm) );
    g_d3dprm.Windowed               = pDeviceInfo->bWindowed;
    g_d3dprm.BackBufferCount        = 1;
    g_d3dprm.MultiSampleType        = pDeviceInfo->MultiSampleType;
    g_d3dprm.SwapEffect             = D3DSWAPEFFECT_DISCARD;
    g_d3dprm.EnableAutoDepthStencil = g_bUseDepthBuffer;
    g_d3dprm.AutoDepthStencilFormat = pModeInfo->DepthStencilFormat;
    g_d3dprm.hDeviceWindow          = hWnd;
    if( g_d3dprm.Windowed )
    {
        g_d3dprm.BackBufferWidth  = g_rcWindowClient.right - g_rcWindowClient.left;
        g_d3dprm.BackBufferHeight = g_rcWindowClient.bottom - g_rcWindowClient.top;
        g_d3dprm.BackBufferFormat = pAdapterInfo->d3ddmDesktop.Format;
		g_bWindowed = true;
	}
    else
    {
        g_d3dprm.BackBufferWidth  = pModeInfo->Width;
        g_d3dprm.BackBufferHeight = pModeInfo->Height;
        g_d3dprm.BackBufferFormat = pModeInfo->Format;
		g_bWindowed = false;
    }

    // Create the device
    hr = g_lpD3D->CreateDevice( g_dwAdapter, pDeviceInfo->DeviceType,
                               hWnd, pModeInfo->dwBehavior, &g_d3dprm,
                               &g_lpD3DDEV );
    if( SUCCEEDED(hr) ){
        if( g_bWindowed )
        {
            SetWindowPos( hWnd, HWND_NOTOPMOST,
                          g_rcWindowBounds.left, g_rcWindowBounds.top,
                          ( g_rcWindowBounds.right - g_rcWindowBounds.left ),
                          ( g_rcWindowBounds.bottom - g_rcWindowBounds.top ),
                          SWP_SHOWWINDOW );
        }
        hr = InitDeviceObjects();
        if( SUCCEEDED(hr) ){
            hr = RestoreDeviceObjects();
            if( SUCCEEDED(hr) ){
                g_bReady = TRUE;
                return S_OK;
            }
        }

        // Cleanup before we try again
        InvalidateDeviceObjects();
        DeleteDeviceObjects();
        SAFE_RELEASE( g_lpD3DDEV );
		return	hr;
	}

    // If that failed, fall back to the reference rasterizer
    if( pDeviceInfo->DeviceType == D3DDEVTYPE_HAL )
    {
        // Let the user know we are switching from HAL to the reference rasterizer
	    TCHAR strMsg[512];
		_tcscat( strMsg, _T("\n\nSwitching to the reference rasterizer,\n")
						_T("a software device that implements the entire\n")
						_T("Direct3D feature set, but runs very slowly.") );
		MessageBox(NULL,strMsg,"Direct3D warning",MB_ICONERROR|MB_OK);

        // Select the default adapter
        g_dwAdapter = 0L;
        pAdapterInfo = &g_Adapters[g_dwAdapter];

        // Look for a software device
        for( UINT i=0L; i<pAdapterInfo->dwNumDevices; i++ )
        {
            if( pAdapterInfo->devices[i].DeviceType == D3DDEVTYPE_REF )
            {
                pAdapterInfo->dwCurrentDevice = i;
                pDeviceInfo = &pAdapterInfo->devices[i];
				g_bWindowed = pDeviceInfo->bWindowed;
                break;
            }
        }

        // Try again, this time with the reference rasterizer
        if( pAdapterInfo->devices[pAdapterInfo->dwCurrentDevice].DeviceType ==
            D3DDEVTYPE_REF )
        {
            hr = Initialize3DEnvironment(hWnd);
        }
    }
    return hr;
}

void	UpdateScreen()
{
	float	timeElapsed;
	HRESULT	hr;
	if (g_pHighResTimer == NULL)
		return;
	if (0.5 > g_pHighResTimer->GetElapsedTime())
		return;
	timeElapsed=g_pHighResTimer->GetElapsedTimeAndReset();
	if (g_pObject3D)
		g_pObject3D->Update(timeElapsed);
    // Test the cooperative level to see if it's okay to render
    if( FAILED( hr = g_lpD3DDEV->TestCooperativeLevel() ) )
    {
        // If the device was lost, do not render until we get it back
        if( D3DERR_DEVICELOST == hr )
            return;

        // Check if the device needs to be resized.
        if( D3DERR_DEVICENOTRESET == hr )
        {
            // If we are windowed, read the desktop mode and use the same format for
            // the back buffer
            if( g_bWindowed )
            {
                D3DAdapterInfo* pAdapterInfo = &g_Adapters[g_dwAdapter];
                g_lpD3D->GetAdapterDisplayMode( g_dwAdapter, &pAdapterInfo->d3ddmDesktop );
                g_d3dprm.BackBufferFormat = pAdapterInfo->d3ddmDesktop.Format;
            }

            if( FAILED( hr = Resize3DEnvironment() ) )
                return;
        }
        return;
    }

	if (g_bUseDepthBuffer)
		g_lpD3DDEV->Clear(0,NULL,D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,g_bgColor,1.0f,0);
	else
		g_lpD3DDEV->Clear(0,NULL,D3DCLEAR_TARGET,g_bgColor,1.0f,0);

	g_lpD3DDEV->BeginScene();
	SetDefaultRenderState(g_lpD3DDEV);
	SetupLights();
	SetDefaultTransform(g_lpD3DDEV);
	if (g_pObject3D)
		g_pObject3D->Render(g_lpD3DDEV);
	g_lpD3DDEV->EndScene();

	if (FAILED(g_lpD3DDEV->Present(NULL,NULL,NULL,NULL))){
		Resize3DEnvironment();
	}
}

void	SetDefaultRenderState(LPDIRECT3DDEVICE8 lpd3ddev)
{
	if (g_bUseDepthBuffer)
		lpd3ddev->SetRenderState(D3DRS_ZENABLE,TRUE);
	else
		lpd3ddev->SetRenderState(D3DRS_ZENABLE,FALSE);
	lpd3ddev->SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);
    lpd3ddev->SetRenderState(D3DRS_DITHERENABLE, TRUE );
	lpd3ddev->SetRenderState(D3DRS_ALPHABLENDENABLE,FALSE);
	lpd3ddev->SetRenderState(D3DRS_BLENDOP,FALSE);
}

void	SetDefaultTransform(LPDIRECT3DDEVICE8 lpd3ddev)
{
	float aspect;
	D3DXMATRIX	matWorld,matRotateX,matRotateZ,matRotateY,matTrans;

	D3DXMatrixLookAtLH(&g_matView,&D3DXVECTOR3(0,0.0f,CAMERA_Z),
                                  &D3DXVECTOR3( 0.0f, 0.0f, 0.0f ), 
                                  &D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) );
	aspect = (float)g_d3dsdBackBuffer.Width / (float)g_d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH( &g_matProj, D3DX_PI/4, aspect, 1.0f, 100.0f );
	lpd3ddev->SetTransform(D3DTS_VIEW,&g_matView);
	lpd3ddev->SetTransform(D3DTS_PROJECTION,&g_matProj);
}

//-----------------------------------------------------------------------------
// Name: SetupLights()
// Desc: Sets up the lights and materials for the scene.
//-----------------------------------------------------------------------------
VOID SetupLights()
{
	HRESULT	hr;
    D3DXVECTOR3 vecDir;
    D3DLIGHT8 light;
    ZeroMemory( &light, sizeof(D3DLIGHT8) );
    light.Type       = D3DLIGHT_DIRECTIONAL;
    light.Diffuse.r  = 1.0f;
    light.Diffuse.g  = 1.0f;
    light.Diffuse.b  = 1.0f;
	vecDir = D3DXVECTOR3(	-1.5,
                        -2.0f,
                        8.0f);
	g_vecLight = vecDir;	//	Copy it.
    D3DXVec3Normalize( (D3DXVECTOR3*)&light.Direction, &vecDir );
    light.Range       = 1000.0f;
    hr = g_lpD3DDEV->SetLight( 0, &light );
    hr = g_lpD3DDEV->LightEnable( 0, TRUE );
    g_lpD3DDEV->SetRenderState( D3DRS_LIGHTING, TRUE );
    g_lpD3DDEV->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_RGBA(96,96,96,255));
}

//-----------------------------------------------------------------------------
// Name:InitDeviceObjects
// Desc:  Initialize device dependent objects.
//-----------------------------------------------------------------------------
HRESULT	InitDeviceObjects()
{
	if(g_pObject3D)
		g_pObject3D->InitDeviceObjects(g_lpD3DDEV);
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:RestoreDeviceObjects
// Desc:  Restore device dependent objects.
//-----------------------------------------------------------------------------
HRESULT	RestoreDeviceObjects()
{
    LPDIRECT3DSURFACE8 pBackBuffer;
    g_lpD3DDEV->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
    pBackBuffer->GetDesc( &g_d3dsdBackBuffer );
    pBackBuffer->Release();
	if(g_pObject3D)
		g_pObject3D->RestoreDeviceObjects(g_lpD3DDEV);
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:InvalidateDeviceObjects
// Desc:  Invalidate device dependent objects.
//-----------------------------------------------------------------------------
HRESULT	InvalidateDeviceObjects()
{
	if(g_pObject3D)
		g_pObject3D->InvalidateDeviceObjects();
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:DeleteDeviceObjects
// Desc:  Delete device dependent objects.
//-----------------------------------------------------------------------------
HRESULT	DeleteDeviceObjects()
{
	if(g_pObject3D)
		g_pObject3D->DeleteDeviceObjects();
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
HRESULT Resize3DEnvironment()
{
    HRESULT hr;

    // Release all vidmem objects
    if( FAILED( hr = InvalidateDeviceObjects() ) )
        return hr;

    g_d3dprm.hDeviceWindow = g_hWnd;
	GetClientRect( g_hWnd, &g_rcWindowClient );
	if( g_bWindowed )
	{
		g_d3dprm.BackBufferWidth  = g_rcWindowClient.right - g_rcWindowClient.left;
		g_d3dprm.BackBufferHeight = g_rcWindowClient.bottom - g_rcWindowClient.top;
	}
    // Reset the device
    if( FAILED( hr = g_lpD3DDEV->Reset( &g_d3dprm ) ) )
        return hr;

    // Initialize the app's device-dependent objects
    hr = RestoreDeviceObjects();
    if( FAILED(hr) )
        return hr;

	if (g_pHighResTimer != NULL)
		g_pHighResTimer->Reset();
    return S_OK;
}



HRESULT	ForceFullscreen()
{
	if (g_bWindowed)
		return	ToggleFullscreen();
	return	S_OK;
}
//-----------------------------------------------------------------------------
// Name: ToggleFullScreen()
// Desc: Called when user toggles between fullscreen mode and windowed mode
//-----------------------------------------------------------------------------
HRESULT ToggleFullscreen()
{
    // Get access to current adapter, device, and mode
    D3DAdapterInfo* pAdapterInfo = &g_Adapters[g_dwAdapter];
    D3DDeviceInfo*  pDeviceInfo  = &pAdapterInfo->devices[pAdapterInfo->dwCurrentDevice];
    D3DModeInfo*    pModeInfo    = &pDeviceInfo->modes[pDeviceInfo->dwCurrentMode];

    // Need device change if going windowed and the current device
    // can only be fullscreen
    if( !g_bWindowed && !pDeviceInfo->bCanDoWindowed )
        return ForceWindowed();
	g_bReady = FALSE;
    // Toggle the windowed state
    g_bWindowed = !g_bWindowed;
	pDeviceInfo->bWindowed = g_bWindowed;

    // Prepare window for windowed/fullscreen change
    AdjustWindowForChange();
    // Set up the presentation parameters
    g_d3dprm.Windowed               = pDeviceInfo->bWindowed;
    g_d3dprm.MultiSampleType        = pDeviceInfo->MultiSampleType;
    g_d3dprm.AutoDepthStencilFormat = pModeInfo->DepthStencilFormat;
    g_d3dprm.hDeviceWindow          = g_hWnd;
    if( g_bWindowed )
    {
        g_d3dprm.BackBufferWidth  = g_rcWindowClient.right - g_rcWindowClient.left;
        g_d3dprm.BackBufferHeight = g_rcWindowClient.bottom - g_rcWindowClient.top;
        g_d3dprm.BackBufferFormat = pAdapterInfo->d3ddmDesktop.Format;
    }
    else
    {
        g_d3dprm.BackBufferWidth  = pModeInfo->Width;
        g_d3dprm.BackBufferHeight = pModeInfo->Height;
        g_d3dprm.BackBufferFormat = pModeInfo->Format;
    }

    // Resize the 3D device
    if( FAILED( Resize3DEnvironment() ) )
    {
        if( g_bWindowed )
            return ForceWindowed();
        else
            return E_FAIL;
    }
	g_bReady = TRUE;
	if (g_pHighResTimer != NULL)
		g_pHighResTimer->Reset();
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: ForceWindowed()
// Desc: Switch to a windowed mode, even if that means picking a new device
//       and/or adapter
//-----------------------------------------------------------------------------
HRESULT ForceWindowed()
{
    HRESULT hr;
    D3DAdapterInfo* pAdapterInfoCur = &g_Adapters[g_dwAdapter];
    D3DDeviceInfo*  pDeviceInfoCur  = &pAdapterInfoCur->devices[pAdapterInfoCur->dwCurrentDevice];
    BOOL bFoundDevice = FALSE;

    if( pDeviceInfoCur->bWindowed )
    {
        bFoundDevice = TRUE;
    }
    else
    {
        // Look for a windowable device on any adapter
        D3DAdapterInfo* pAdapterInfo;
        DWORD dwAdapter;
        D3DDeviceInfo* pDeviceInfo;
        DWORD dwDevice;
        for( dwAdapter = 0; dwAdapter < g_dwNumAdapters; dwAdapter++ )
        {
            pAdapterInfo = &g_Adapters[dwAdapter];
            for( dwDevice = 0; dwDevice < pAdapterInfo->dwNumDevices; dwDevice++ )
            {
                pDeviceInfo = &pAdapterInfo->devices[dwDevice];
                if( pDeviceInfo->bWindowed )
                {
                    g_dwAdapter = dwAdapter;
                    pDeviceInfoCur = pDeviceInfo;
                    pAdapterInfo->dwCurrentDevice = dwDevice;
                    bFoundDevice = TRUE;
                    break;
                }
            }
            if( bFoundDevice )
                break;
        }
    }

    if( !bFoundDevice )
        return E_FAIL;

    pDeviceInfoCur->bWindowed = TRUE;
    g_bWindowed = TRUE;
	g_bReady = FALSE;
    // Release all scene objects that will be re-created for the new device
    InvalidateDeviceObjects();
    DeleteDeviceObjects();

    // Release display objects, so a new device can be created
    if( g_lpD3DDEV->Release() > 0L )
        return E_FAIL;
        //	DisplayErrorMsg( D3DAPPERR_NONZEROREFCOUNT, MSGERR_APPMUSTEXIT );

    // Create the new device
    if( FAILED( hr = Initialize3DEnvironment(g_hWnd) ) )
        return E_FAIL;
        //	DisplayErrorMsg( hr, MSGERR_APPMUSTEXIT );
    g_bReady = TRUE;
	if (g_pHighResTimer != NULL)
		g_pHighResTimer->Reset();
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: AdjustWindowForChange()
// Desc: Prepare the window for a possible change between windowed mode and
//       fullscreen mode.  This function is virtual and thus can be overridden
//       to provide different behavior, such as switching to an entirely
//       different window for fullscreen mode (as in the MFC sample apps).
//-----------------------------------------------------------------------------
HRESULT AdjustWindowForChange()
{
    if( g_bWindowed )
    {
        // Set windowed-mode style
        SetWindowLong( g_hWnd, GWL_STYLE, g_dwWindowStyle );
        SetWindowPos( g_hWnd, HWND_NOTOPMOST,
                      g_rcWindowBounds.left, g_rcWindowBounds.top,
                      ( g_rcWindowBounds.right - g_rcWindowBounds.left ),
                      ( g_rcWindowBounds.bottom - g_rcWindowBounds.top ),
                      SWP_SHOWWINDOW|SWP_DRAWFRAME );
    }
    else
    {
        // Set fullscreen-mode style
        SetWindowLong( g_hWnd, GWL_STYLE, WS_POPUP|WS_SYSMENU|WS_VISIBLE );
		HWND	hWnd;
		RECT	rcScreen;
		hWnd = ::GetDesktopWindow();
		if (::GetWindowRect(hWnd,&rcScreen)){
			SetWindowPos( g_hWnd, HWND_NOTOPMOST,
                      rcScreen.left, rcScreen.top,
                      ( rcScreen.right - rcScreen.left ),
                      ( rcScreen.bottom - rcScreen.top ),
                      SWP_SHOWWINDOW );
		}
	}
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: UserSelectNewDevice()
// Desc: Displays a dialog so the user can select a new adapter, device, or
//       display mode, and then recreates the 3D environment if needed
//-----------------------------------------------------------------------------
HRESULT UserSelectNewDevice()
{
    return S_OK;
}
//-----------------------------------------------------------------------------
// Name: UserAdjustWindow()
// Desc: Called when user changed the window position and size.
//-----------------------------------------------------------------------------
HRESULT UserAdjustWindow(){
	if (!g_bWindowed)
		return S_OK;
	::GetWindowRect(g_hWnd,&g_rcWindowBounds);
	::GetClientRect(g_hWnd,&g_rcWindowClient);
	return Resize3DEnvironment();
}
void	AdjustCursor(){
	if (g_bWindowed)
		::SetCursor(::LoadCursor(NULL,IDC_ARROW));
	else
		::SetCursor(NULL);
}

