//
//	
//
//
//
#include "stdafx.h"
#include "Resource.h"
#include <d3d8.h>
#include <d3dx8.h>
#include <dxfile.h>
#include "GameMain.h"
#include "XFrame.h"
#include "Object3D.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif
#define	MAX_MESH_RADIUS	2.0f

//
// \z/
//

CObject3D::CObject3D()
{
	m_pCXFrame = NULL;
	m_bUseHWNPatches = FALSE;
	m_dwNumSegs = 0;
	m_fAngleY = 0;
}

CObject3D::~CObject3D()
{
	this->InvalidateDeviceObjects();
	this->DeleteDeviceObjects();
}

void CObject3D::SetMeshFilename(TCHAR *path, TCHAR *fname){
}



//===================================================================
// Name: InitDeviceObjects()
// Desc: foCXˑIuWFNg̍\ziVXej
//===================================================================
HRESULT	CObject3D::InitDeviceObjects(LPDIRECT3DDEVICE8 lpd3ddev)
{
	InvalidateDeviceObjects();
	DeleteDeviceObjects();

	m_pCXFrame = new CXFrame();
	DXFILELOADRESOURCE res;

	res.hModule=GetModuleHandle(NULL);
	res.lpName=MAKEINTRESOURCE(IDR_XFILE1);
	res.lpType="xfile";
	m_pCXFrame->InitDeviceObjects(lpd3ddev,&res);

	m_dwNumMesh = m_pCXFrame->CountNMeshes();
	TRACE("Total %d meshes\n",m_dwNumMesh);
	m_dwNumFrame = m_pCXFrame->CountNFrames();
	TRACE("Total %d frames\n",m_dwNumFrame);

	D3DXMATRIX	matIdentity;
	D3DXMatrixIdentity(&matIdentity);
	m_pCXFrame->AdjustPositions(lpd3ddev,&matIdentity);
	D3DXMatrixScaling(&m_matScale,1.0f,1.0f,1.0f);
	m_fScale = 1.0f;

    return S_OK;
}


//===================================================================
// Name: RestoreDeviceObjects()
// Desc: 
//===================================================================
HRESULT CObject3D::RestoreDeviceObjects(LPDIRECT3DDEVICE8 lpd3ddev)
{
    HRESULT    hr;
	D3DCAPS8	d3dCaps;

	if (!m_pCXFrame)
		return	E_FAIL;
	InvalidateDeviceObjects();

	lpd3ddev->GetDeviceCaps(&d3dCaps);
	m_bUseHWNPatches = (d3dCaps.DevCaps & D3DDEVCAPS_NPATCHES);
	if (m_pCXFrame){
		hr = m_pCXFrame->RestoreFrameDeviceObjects(lpd3ddev,this->m_dwNumSegs,this->m_bUseHWNPatches);
	}
    if( FAILED(hr) )
        goto	eFail;
    return S_OK;
eFail:
	InvalidateDeviceObjects();
	return	hr;
}


//===================================================================
// Name: InvalidateDeviceObjects()
// Desc: 
//===================================================================
HRESULT	CObject3D::InvalidateDeviceObjects()
{
	if (m_pCXFrame)
		m_pCXFrame->InvalidateFrameDeviceObjects();
	return	S_OK;
}

//===================================================================
// Name: DeleteDeviceObjects()
// Desc: 
//===================================================================
HRESULT	CObject3D::DeleteDeviceObjects()
{
	if (m_pCXFrame != NULL){
		delete m_pCXFrame;
		m_pCXFrame = NULL;
	}
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name: SetScale()
// Desc: XP[O̐ݒ
//-----------------------------------------------------------------------------
void	CObject3D::SetScale(float s){
	m_fScale = s;
	D3DXMatrixScaling(&m_matScale,s,s,s);
}

void	CObject3D::Update(float timeElapsed)
{
    D3DXMATRIX    matTrans, matRotateY;
    m_fAngleY += timeElapsed * (D3DX_PI / 240.0f);
    if (m_fAngleY > D3DX_PI*2.0f)
        m_fAngleY -= D3DX_PI*2.0f;
    D3DXMatrixRotationY(&matRotateY,m_fAngleY);
    D3DXMatrixTranslation(&matTrans,0.0f,0.0f,2.0f);
    m_matWorld =  matRotateY * matTrans;
}


void	CObject3D::Render(LPDIRECT3DDEVICE8 lpd3ddev)
{
	lpd3ddev->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	lpd3ddev->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	lpd3ddev->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
	lpd3ddev->SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG2);
	lpd3ddev->SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);

	lpd3ddev->SetRenderState(D3DRS_ALPHABLENDENABLE,FALSE);
    lpd3ddev->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
    lpd3ddev->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	lpd3ddev->SetRenderState(D3DRS_LIGHTING,TRUE);
	this->SetupLights();
    if (m_bUseHWNPatches)
    {
		float fNumSegs;
		fNumSegs = (float)m_dwNumSegs;
		lpd3ddev->SetRenderState(D3DRS_PATCHSEGMENTS, *((DWORD*)&fNumSegs));
	}

	D3DXMATRIX	matIdentity, matTmp;
	D3DXMatrixIdentity(&matIdentity);
	matTmp = m_matScale * m_matWorld;
	m_pCXFrame->RenderFrames(lpd3ddev,&matTmp,&matIdentity);

    if (m_bUseHWNPatches)
    {
        lpd3ddev->SetRenderState(D3DRS_PATCHSEGMENTS, 0);
    }
}

void CObject3D::SetupLights()
{
	HRESULT	hr;
    D3DXVECTOR3 vecDir;
    D3DLIGHT8 light;
    ZeroMemory( &light, sizeof(D3DLIGHT8) );
    light.Type       = D3DLIGHT_DIRECTIONAL;
    light.Diffuse.r  = 1.0f * m_fScale;
    light.Diffuse.g  = 1.0f * m_fScale;
    light.Diffuse.b  = 1.0f * m_fScale;
	vecDir = D3DXVECTOR3(	-1.5,
                        -2.0f,
                        8.0f);
    D3DXVec3Normalize( (D3DXVECTOR3*)&light.Direction, &vecDir );
    light.Range       = 1000.0f;
    hr = g_lpD3DDEV->SetLight( 0, &light );
    hr = g_lpD3DDEV->LightEnable( 0, TRUE );
    g_lpD3DDEV->SetRenderState( D3DRS_LIGHTING, TRUE );
    g_lpD3DDEV->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_RGBA(192,192,192,255));
}
