/*
 *		Basic Direct3D program.
 *		This program shows you an example about 
 *		the typical Initialization and typical 
 *		termination for Direct3D.
 *
 *		First edition: August.12.2002
 */

#include "StdAfx.h"
#include "Resource.h"
#include <math.h>
#include <d3d8.h>
#include <d3dx8.h>
#include <mmsystem.h>
#include "d3dmain.h"
#include "DeviceInformation.h"
#include "HighResTimer.h"
#include <stdio.h>
#include <stdarg.h>
#include <tchar.h>
 /*
 *		Globals.
 */
HWND					g_hWnd;
D3DPRESENT_PARAMETERS	g_d3dprm;
LPDIRECT3D8				g_lpD3D = NULL;
LPDIRECT3DDEVICE8		g_lpD3DDEV = NULL;

CDeviceInformation	*g_lpDevices = NULL;
D3DAdapterInfo		*g_Adapters = NULL;
DWORD				g_dwNumAdapters	= 0;
DWORD				g_dwAdapter		= 0L;
BOOL				g_bWindowed = FALSE;
BOOL				g_bUseDepthBuffer   = TRUE;
DWORD				g_dwMinDepthBits    = 16;
DWORD				g_dwMinStencilBits  = 0;
DWORD				g_dwWindowStyle;
RECT				g_rcWindowBounds;
RECT				g_rcWindowClient;
BOOLEAN				g_bReady = false;
D3DCAPS8			g_CurrentDeviceCaps;
D3DSURFACE_DESC		g_d3dsdBackBuffer;
BOOL				g_bCanGreaterEqual;

LPD3DXFONT			g_pFont = NULL;
LPDIRECT3DTEXTURE8	g_pTexture = NULL;
float			g_fTexImageWidth;
float			g_fTexImageHeight;

D3DXMATRIX		g_matView;
D3DXMATRIX		g_matProj;
D3DXVECTOR3		g_vecLight;
D3DCOLOR		g_bgColor = D3DCOLOR_XRGB(BGC_R,BGC_G,BGC_B);

CHighResTimer	*g_pHighResTimer = NULL;

/*
 *		Constants.
 */
#define	CAMERA_Z	(-5.0f)

/*
 *		Prototypes.
 */
HRESULT Initialize3DEnvironment(HWND hWnd);
void	SetDefaultRenderState(LPDIRECT3DDEVICE8 lpd3ddev);

void	ReleaseD3D()
{
	InvalidateDeviceObjects();
	DeleteDeviceObjects();
	if (g_pHighResTimer != NULL){
		delete g_pHighResTimer;
		g_pHighResTimer = NULL;
	}
	SAFE_RELEASE(g_lpD3DDEV);
	SAFE_RELEASE(g_lpD3D);
	if (g_lpDevices){
		delete	g_lpDevices;
		g_lpDevices = NULL;
	}
}

#define TITLE               "Direct3D s"

//-----------------------------------------------------------------------------
// Name: InitFail()
// Desc: A[g\AAvP[VI
//-----------------------------------------------------------------------------
void	InitFail(HWND hWnd, LPCTSTR szError,...)
{
    char                        szBuff[128];
    va_list                     vl;

    va_start(vl, szError);
    vsprintf(szBuff, szError, vl);
    MessageBox(hWnd, szBuff, TITLE, MB_OK);
    ReleaseD3D();
	DestroyWindow(hWnd);
    va_end(vl);
    return;
}

//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Direct3D ̏
//-----------------------------------------------------------------------------
void	InitD3D(HWND hWnd)
{
	g_hWnd = hWnd;
	g_lpD3D = Direct3DCreate8(D3D_SDK_VERSION);
	if (!g_lpD3D){
		InitFail(hWnd,"Failed to create Direct3D");
		return;
	}
	g_dwWindowStyle = GetWindowLong( hWnd, GWL_STYLE );
    GetWindowRect( hWnd, &g_rcWindowBounds );
    GetClientRect( hWnd, &g_rcWindowClient );

	g_lpDevices = new CDeviceInformation(g_lpD3D, g_bUseDepthBuffer,
									g_dwMinDepthBits, g_dwMinStencilBits,
									NULL);
    if ((NULL == g_lpDevices)||( NULL == (g_Adapters = g_lpDevices->getAdapters())))
    {
		InitFail(hWnd, "BuildDeviceList Failed.");
        return;
    }
	g_dwNumAdapters = g_lpDevices->getNumAdapters();
	g_dwAdapter = g_lpDevices->getCurrentAdapter();

	//	foCX̐
	if (FAILED(::Initialize3DEnvironment(hWnd))){
		InitFail(hWnd, "CreateDevice Failed.");
		return;
	}
	if (FAILED(g_lpD3DDEV->GetDeviceCaps(&g_CurrentDeviceCaps))){
		InitFail(hWnd, "GetDeviceCaps Failed.");
		return;
	}
	g_bCanGreaterEqual = false;
	if (g_CurrentDeviceCaps.AlphaCmpCaps & D3DPCMPCAPS_GREATEREQUAL)
		g_bCanGreaterEqual = true;

	g_pHighResTimer = new CHighResTimer();
}


//-----------------------------------------------------------------------------
// Name: Initialize3DEnvironment()
// Desc:
//-----------------------------------------------------------------------------
HRESULT Initialize3DEnvironment(HWND hWnd)
{
    HRESULT hr;

    D3DAdapterInfo* pAdapterInfo = &g_Adapters[g_dwAdapter];
    D3DDeviceInfo*  pDeviceInfo  = &pAdapterInfo->devices[pAdapterInfo->dwCurrentDevice];
    D3DModeInfo*    pModeInfo    = &pDeviceInfo->modes[pDeviceInfo->dwCurrentMode];

    //	foCXp[^̐ݒ
    ZeroMemory( &g_d3dprm, sizeof(g_d3dprm) );
    g_d3dprm.Windowed               = pDeviceInfo->bWindowed;
    g_d3dprm.BackBufferCount        = 1;
    g_d3dprm.MultiSampleType        = pDeviceInfo->MultiSampleType;
    g_d3dprm.SwapEffect             = D3DSWAPEFFECT_DISCARD;
    g_d3dprm.EnableAutoDepthStencil = g_bUseDepthBuffer;
    g_d3dprm.AutoDepthStencilFormat = pModeInfo->DepthStencilFormat;
    g_d3dprm.hDeviceWindow          = hWnd;
    if( g_d3dprm.Windowed )
    {
        g_d3dprm.BackBufferWidth  = g_rcWindowClient.right - g_rcWindowClient.left;
        g_d3dprm.BackBufferHeight = g_rcWindowClient.bottom - g_rcWindowClient.top;
        g_d3dprm.BackBufferFormat = pAdapterInfo->d3ddmDesktop.Format;
		g_bWindowed = true;
	}
    else
    {
        g_d3dprm.BackBufferWidth  = pModeInfo->Width;
        g_d3dprm.BackBufferHeight = pModeInfo->Height;
        g_d3dprm.BackBufferFormat = pModeInfo->Format;
		g_bWindowed = false;
    }

    //	w肳ꂽp[^ɂfoCX
    hr = g_lpD3D->CreateDevice( g_dwAdapter, pDeviceInfo->DeviceType,
                               hWnd, pModeInfo->dwBehavior, &g_d3dprm,
                               &g_lpD3DDEV );
    if( SUCCEEDED(hr) ){
        if( g_bWindowed )
        {
            SetWindowPos( hWnd, HWND_NOTOPMOST,
                          g_rcWindowBounds.left, g_rcWindowBounds.top,
                          ( g_rcWindowBounds.right - g_rcWindowBounds.left ),
                          ( g_rcWindowBounds.bottom - g_rcWindowBounds.top ),
                          SWP_SHOWWINDOW );
        }
        hr = InitDeviceObjects();
        if( SUCCEEDED(hr) ){
            hr = RestoreDeviceObjects();
            if( SUCCEEDED(hr) ){
				g_bReady = TRUE;
				return S_OK;
            }
        }

        //	Ďsׁ̈AfoCXˑIuWFNgJ
        InvalidateDeviceObjects();
        DeleteDeviceObjects();
        SAFE_RELEASE( g_lpD3DDEV );
		return	hr;
	}

    //	REF X^CU֕ύX
    if( pDeviceInfo->DeviceType == D3DDEVTYPE_HAL )
    {
        //	bZ[W\
	    TCHAR strMsg[512];
		_tcscpy( strMsg, _T("\tgEFAX^CUgp܂B\n")
						_T("\tgEFAX^CUł́A삪\n")
						_T("ᑬɂȂ܂B") );
		MessageBox(NULL,strMsg,"Direct3D warning",MB_ICONERROR|MB_OK);

        //	A_v^̑I
        g_dwAdapter = 0L;
        pAdapterInfo = &g_Adapters[g_dwAdapter];

        //	gp\ȃA_v^̌
        for( UINT i=0L; i<pAdapterInfo->dwNumDevices; i++ )
        {
            if( pAdapterInfo->devices[i].DeviceType == D3DDEVTYPE_REF )
            {
                pAdapterInfo->dwCurrentDevice = i;
                pDeviceInfo = &pAdapterInfo->devices[i];
				g_bWindowed = pDeviceInfo->bWindowed;
                break;
            }
        }

        //	REF X^CUw肵čĐݒ
		if( pAdapterInfo->devices[pAdapterInfo->dwCurrentDevice].DeviceType ==
            D3DDEVTYPE_REF )
        {
            hr = Initialize3DEnvironment(hWnd);
        }
    }
    return hr;
}

void	UpdateScreen()
{
	float	timeElapsed;
	HRESULT	hr;
	if (g_pHighResTimer == NULL)
		return;
	if (0.5 > g_pHighResTimer->GetElapsedTime())
		return;
	timeElapsed=g_pHighResTimer->GetElapsedTimeAndReset();
    //	_[\ǂeXg
    if( FAILED( hr = g_lpD3DDEV->TestCooperativeLevel() ) )
    {
        //	D3DERR_DEVICELOST Ȃ΁A_[͍sȂ
        if( D3DERR_DEVICELOST == hr )
            return;

        //	foCX̃ZbgKvȏꍇ.
        if( D3DERR_DEVICENOTRESET == hr )
        {
            //	EBhE[hł΃fXNgbvŎgpĂ郂[h
            //	킹B
            if( g_bWindowed )
            {
                D3DAdapterInfo* pAdapterInfo = &g_Adapters[g_dwAdapter];
                g_lpD3D->GetAdapterDisplayMode( g_dwAdapter, &pAdapterInfo->d3ddmDesktop );
                g_d3dprm.BackBufferFormat = pAdapterInfo->d3ddmDesktop.Format;
            }

            if( FAILED( hr = Resize3DEnvironment() ) )
                return;
        }
        return;
    }

	if (g_bUseDepthBuffer)
		g_lpD3DDEV->Clear(0,NULL,D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,g_bgColor,1.0f,0);
	else
		g_lpD3DDEV->Clear(0,NULL,D3DCLEAR_TARGET,g_bgColor,1.0f,0);

	g_lpD3DDEV->BeginScene();
	SetDefaultRenderState(g_lpD3DDEV);

	g_lpD3DDEV->EndScene();

	if (FAILED(g_lpD3DDEV->Present(NULL,NULL,NULL,NULL))){
		Resize3DEnvironment();
	}
}

void	SetDefaultRenderState(LPDIRECT3DDEVICE8 lpd3ddev)
{
	lpd3ddev->SetRenderState(D3DRS_ZENABLE,FALSE);
	lpd3ddev->SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);
    lpd3ddev->SetRenderState(D3DRS_DITHERENABLE, TRUE );
	lpd3ddev->SetRenderState(D3DRS_BLENDOP,FALSE);
    lpd3ddev->LightEnable( 0, FALSE );
	lpd3ddev->SetRenderState( D3DRS_LIGHTING, FALSE );
    g_lpD3DDEV->SetTextureStageState( 0,D3DTSS_COLOROP, D3DTOP_MODULATE );
    g_lpD3DDEV->SetTextureStageState( 0,D3DTSS_COLORARG1,D3DTA_TEXTURE );
    g_lpD3DDEV->SetTextureStageState( 0,D3DTSS_COLORARG2,D3DTA_DIFFUSE );
    g_lpD3DDEV->SetTextureStageState( 0,D3DTSS_MIPFILTER,D3DTEXF_LINEAR );
    g_lpD3DDEV->SetTextureStageState( 0,D3DTSS_MAGFILTER,D3DTEXF_LINEAR );
    g_lpD3DDEV->SetTextureStageState( 0,D3DTSS_MINFILTER,D3DTEXF_LINEAR );
}


//-----------------------------------------------------------------------------
// Name:InitDeviceObjects
// Desc:  foCXˑIuWFNg̏
//-----------------------------------------------------------------------------
HRESULT	InitDeviceObjects()
{
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:RestoreDeviceObjects
// Desc:  rfIJ[h̃foCXˑIuWFNg𐶐
//-----------------------------------------------------------------------------
HRESULT	RestoreDeviceObjects()
{
    LPDIRECT3DSURFACE8 pBackBuffer;
    g_lpD3DDEV->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
    pBackBuffer->GetDesc( &g_d3dsdBackBuffer );
    pBackBuffer->Release();
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:InvalidateDeviceObjects
// Desc:  rfIJ[h̃foCXˑIuWFNgj
//-----------------------------------------------------------------------------
HRESULT	InvalidateDeviceObjects()
{
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:DeleteDeviceObjects
// Desc:  foCXˑIuWFNg̍폜
//-----------------------------------------------------------------------------
HRESULT	DeleteDeviceObjects()
{
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
HRESULT Resize3DEnvironment()
{
    HRESULT hr;

    //	rfIJ[h̃foCXˑIuWFNgj
    if( FAILED( hr = InvalidateDeviceObjects() ) )
        return hr;

    g_d3dprm.hDeviceWindow = g_hWnd;
	GetClientRect( g_hWnd, &g_rcWindowClient );
	if( g_bWindowed )
	{
		g_d3dprm.BackBufferWidth  = g_rcWindowClient.right - g_rcWindowClient.left;
		g_d3dprm.BackBufferHeight = g_rcWindowClient.bottom - g_rcWindowClient.top;
	}
    //	`foCXZbg
    if( FAILED( hr = g_lpD3DDEV->Reset( &g_d3dprm ) ) )
        return hr;

    //	rfIJ[h̃foCXˑIuWFNg𐶐
    hr = RestoreDeviceObjects();
    if( FAILED(hr) )
        return hr;

	if (g_pHighResTimer != NULL)
		g_pHighResTimer->Reset();
    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: ForceFullscreen()
// Desc: \tXN[ɂB
//-----------------------------------------------------------------------------
HRESULT	ForceFullscreen()
{
	if (g_bWindowed)
		return	ToggleFullscreen();
	return	S_OK;
}

//-----------------------------------------------------------------------------
// Name: ToggleFullScreen()
// Desc: tXN[EEBhE[h̐؂ւ
//-----------------------------------------------------------------------------
HRESULT ToggleFullscreen()
{
    //	gp̃foCX擾
    D3DAdapterInfo* pAdapterInfo = &g_Adapters[g_dwAdapter];
    D3DDeviceInfo*  pDeviceInfo  = &pAdapterInfo->devices[pAdapterInfo->dwCurrentDevice];
    D3DModeInfo*    pModeInfo    = &pDeviceInfo->modes[pDeviceInfo->dwCurrentMode];

    //	EChE[hɈڍŝAgp̃foCX
    //	tXN[pƂ̑Ώ
    if( !g_bWindowed && !pDeviceInfo->bCanDoWindowed )
        return ForceWindowed();
	g_bReady = FALSE;
    //	؂ւ
    g_bWindowed = !g_bWindowed;
	pDeviceInfo->bWindowed = g_bWindowed;

    //	Windows ̃EBhEX^C𒲐
    AdjustWindowForChange();
    //	p[^̐ݒ
    g_d3dprm.Windowed               = pDeviceInfo->bWindowed;
    g_d3dprm.MultiSampleType        = pDeviceInfo->MultiSampleType;
    g_d3dprm.AutoDepthStencilFormat = pModeInfo->DepthStencilFormat;
    g_d3dprm.hDeviceWindow          = g_hWnd;
    if( g_bWindowed )
    {
        g_d3dprm.BackBufferWidth  = g_rcWindowClient.right - g_rcWindowClient.left;
        g_d3dprm.BackBufferHeight = g_rcWindowClient.bottom - g_rcWindowClient.top;
        g_d3dprm.BackBufferFormat = pAdapterInfo->d3ddmDesktop.Format;
    }
    else
    {
        g_d3dprm.BackBufferWidth  = pModeInfo->Width;
        g_d3dprm.BackBufferHeight = pModeInfo->Height;
        g_d3dprm.BackBufferFormat = pModeInfo->Format;
    }

    // foCX̃TCY
    if( FAILED( Resize3DEnvironment() ) )
    {
        if( g_bWindowed )
            return ForceWindowed();
        else
            return E_FAIL;
    }
	g_bReady = TRUE;
	if (g_pHighResTimer != NULL)
		g_pHighResTimer->Reset();
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: ForceWindowed()
// Desc: EBhE[hɐ؂ւEKvȂ\[h؂ւ
//-----------------------------------------------------------------------------
HRESULT ForceWindowed()
{
    HRESULT hr;
    D3DAdapterInfo* pAdapterInfoCur = &g_Adapters[g_dwAdapter];
    D3DDeviceInfo*  pDeviceInfoCur  = &pAdapterInfoCur->devices[pAdapterInfoCur->dwCurrentDevice];
    BOOL bFoundDevice = FALSE;

    if( pDeviceInfoCur->bWindowed )
    {
        bFoundDevice = TRUE;
    }
    else
    {
        //	EBhE`\ȃfoCXT
        D3DAdapterInfo* pAdapterInfo;
        DWORD dwAdapter;
        D3DDeviceInfo* pDeviceInfo;
        DWORD dwDevice;
        for( dwAdapter = 0; dwAdapter < g_dwNumAdapters; dwAdapter++ )
        {
            pAdapterInfo = &g_Adapters[dwAdapter];
            for( dwDevice = 0; dwDevice < pAdapterInfo->dwNumDevices; dwDevice++ )
            {
                pDeviceInfo = &pAdapterInfo->devices[dwDevice];
                if( pDeviceInfo->bWindowed )
                {
                    g_dwAdapter = dwAdapter;
                    pDeviceInfoCur = pDeviceInfo;
                    pAdapterInfo->dwCurrentDevice = dwDevice;
                    bFoundDevice = TRUE;
                    break;
                }
            }
            if( bFoundDevice )
                break;
        }
    }

    if( !bFoundDevice )
        return E_FAIL;

    pDeviceInfoCur->bWindowed = TRUE;
    g_bWindowed = TRUE;
	g_bReady = FALSE;
    // foCXˑIuWFNg̔j
    InvalidateDeviceObjects();
    DeleteDeviceObjects();

    // foCX̔j
    if( g_lpD3DDEV->Release() > 0L )
        return E_FAIL;
        //	DisplayErrorMsg( D3DAPPERR_NONZEROREFCOUNT, MSGERR_APPMUSTEXIT );

    // VfoCX̐
    if( FAILED( hr = Initialize3DEnvironment(g_hWnd) ) )
        return E_FAIL;
        //	DisplayErrorMsg( hr, MSGERR_APPMUSTEXIT );
    g_bReady = TRUE;

	//	^C}[Zbg
	if (g_pHighResTimer != NULL)
		g_pHighResTimer->Reset();
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: AdjustWindowForChange()
// Desc: tXN[[hEEChE[h؂ւ
//		EBhEX^Ĉ܂킹B
//-----------------------------------------------------------------------------
HRESULT AdjustWindowForChange()
{
    if( g_bWindowed )
    {
        // Set windowed-mode style
        SetWindowLong( g_hWnd, GWL_STYLE, g_dwWindowStyle );
        SetWindowPos( g_hWnd, HWND_NOTOPMOST,
                      g_rcWindowBounds.left, g_rcWindowBounds.top,
                      ( g_rcWindowBounds.right - g_rcWindowBounds.left ),
                      ( g_rcWindowBounds.bottom - g_rcWindowBounds.top ),
                      SWP_SHOWWINDOW|SWP_DRAWFRAME );
    }
    else
    {
        // Set fullscreen-mode style
        SetWindowLong( g_hWnd, GWL_STYLE, WS_POPUP|WS_SYSMENU|WS_VISIBLE );
		HWND	hWnd;
		RECT	rcScreen;
		hWnd = ::GetDesktopWindow();
		if (::GetWindowRect(hWnd,&rcScreen)){
			SetWindowPos( g_hWnd, HWND_NOTOPMOST,
                      rcScreen.left, rcScreen.top,
                      ( rcScreen.right - rcScreen.left ),
                      ( rcScreen.bottom - rcScreen.top ),
                      SWP_SHOWWINDOW );
		}
	}
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: UserSelectNewDevice()
// Desc: [UɃfoCXI鏈ǉ
//-----------------------------------------------------------------------------
HRESULT UserSelectNewDevice()
{
    return S_OK;
}
//-----------------------------------------------------------------------------
// Name: UserAdjustWindow()
// Desc: EChETCY̕ύXȂ.
//-----------------------------------------------------------------------------
HRESULT UserAdjustWindow(){
	if (!g_bWindowed)
		return S_OK;
	::GetWindowRect(g_hWnd,&g_rcWindowBounds);
	::GetClientRect(g_hWnd,&g_rcWindowClient);
	return Resize3DEnvironment();
}

//-----------------------------------------------------------------------------
// Name: AdjustCursor()
// Desc: tXN[ɃJ[\
//-----------------------------------------------------------------------------
void	AdjustCursor(){
	if (g_bWindowed)
		::SetCursor(::LoadCursor(NULL,IDC_ARROW));
	else
		::SetCursor(NULL);
}

